from typing import List, Union

import torch

from esp_ppq.core import OperationQuantizationConfig, SingletonMeta

ACTIVATION_OP_SET = {
    "Relu",
    "PRelu",
    "Sigmoid",
    "Tanh",
    "HardSwish",
    "HardSigmoid",
    "Elu",
    "Erf",
    "Gelu",
    "Clip",
    "Cast",
    "LeakyRelu",
    "Softplus",
    "Identity",
    "Swish",
}

MATH_OP_SET = {
    "Abs",
    "Exp",
    "Log",
    "Sqrt",
    "Cos",
    "Sin",
    "Tan",
    "Not",
    "Floor",
    "Round",
    "Ceil",
    "Neg",
}

QUANT_OP_SET = {
    "RequantizeLinear",
    "QuantizeLinear",
    "DequantizeLinear",
    "QuantizeFloating",
    "DequantizeFloating",
}
PASSIVE_LAYOUT_OP_SET = ACTIVATION_OP_SET | QUANT_OP_SET | MATH_OP_SET
CONV_LAYOUT_OP_SET = {
    "Conv",
    "GlobalAveragePool",
    "AveragePool",
    "MaxPool",
    "ConvTranspose",
    "GlobalMaxPool",
    "DepthToSpace",
    "SpaceToDepth",
    "InsertZeros",
}
ADD_LIKE_OP_SET = {
    "Add",
    "Sub",
    "Mul",
    "Div",
    "And",
    "Equal",
    "Greater",
    "Less",
    "GreaterOrEqual",
    "LessOrEqual",
    "Pow",
}
OTHER_OP_SET = {
    "MatMul",
    "Gemm",
    "Flatten",
    "Reshape",
    "Squeeze",
    "Unsqueeze",
    "Transpose",
    "Slice",
    "Pad",
    "Concat",
    "Constant",
    "Gather",
    "GatherElements",
    "GatherND",
    "Shape",
    "ConstantOfShape",
    "ReverseSequence",
    "Expand",
    "LayerNorm",
    "LayerNormalization",
    "Max",
    "Min",
    "ScatterElements",
    "ScatterND",
    "TopK",
    "Where",
    "GRU",
    "LSTM",
    "Sum",
}

SOFTMAX_LIKE_OP_SET = {"Softmax", "LogSoftmax", "Split"}
REDUCE_OP_SET = {
    "ReduceL1",
    "ReduceL2",
    "ReduceMin",
    "ReduceMax",
    "ReduceProd",
    "ReduceSum",
    "ReduceMean",
    "ReduceSumSquare",
    "ReduceLogSum",
    "ReduceLogSumExp",
}
AXIS_TRANSFORM_OP_SET = SOFTMAX_LIKE_OP_SET | REDUCE_OP_SET
# QUANT_EXCLUDE_OP_SET refers to operators that do not participate
# in the operations of quantize, dequantize, or requantize.
QUANT_EXCLUDE_OP_SET = {"Shape"}


class EspQuantType:
    F32 = "F32"
    S16 = "S16"
    S8 = "S8"


class LayoutAnnotation:
    NCHW = "NCHW"
    NHWC = "NHWC"
    HWCN = "HWCN"
    N16HWC16 = "(N/16)HWC16"
    N8HWC8 = "(N/8)HWC8"
    N16HWC16_UNALIGNED = "(N/16)HWC16_UNALIGNED"
    N8HWC8_UNALIGNED = "(N/8)HWC8_UNALIGNED"
    NCW = "NCW"
    NWC = "NWC"
    WCN = "WCN"
    N16WC16 = "(N/16)WC16"
    N8WC8 = "(N/8)WC8"
    N16WC16_UNALIGNED = "(N/16)WC16_UNALIGNED"
    N8WC8_UNALIGNED = "(N/8)WC8_UNALIGNED"
    UNKNOWN = "UNK"


class ExporterPatternInfo(metaclass=SingletonMeta):
    var_exponents = {}
    var_layout = {}
    var_permute = {}
    var_config = {}
    luts = {}

    def reset(self):
        self.var_exponents = {}
        self.var_layout = {}
        self.var_permute = {}
        self.var_config = {}
        self.luts = {}

    def get_var_exponents(self, var_name: str, default: List[int] = None) -> Union[int, List[int]]:
        return self.var_exponents.get(var_name, default)

    def get_var_layout(self, var_name: str, default: LayoutAnnotation = None) -> LayoutAnnotation:
        return self.var_layout.get(var_name, default)

    def get_var_permute(self, var_name: str, default: List[int] = None) -> List[int]:
        return self.var_permute.get(var_name, default)

    def get_var_config(self, var_name: str, default: OperationQuantizationConfig = None) -> OperationQuantizationConfig:
        return self.var_config.get(var_name, default)

    def get_lut(self, lut_name: str, default: torch.Tensor = None) -> torch.Tensor:
        return self.luts.get(lut_name, default)

    def add_var_exponents(self, var_name: str, exponent: Union[int, List[int]]):
        if isinstance(exponent, int):
            exponent = [exponent]
        self.var_exponents[var_name] = exponent

    def add_var_layout(self, var_name: str, layout: LayoutAnnotation):
        self.var_layout[var_name] = layout

    def add_var_permute(self, var_name: str, perm: List[int]):
        self.var_permute[var_name] = perm

    def add_var_config(self, var_name: str, config: OperationQuantizationConfig):
        self.var_config[var_name] = config

    def add_lut(self, lut_name: str, lut: torch.Tensor, exponent: Union[int, List[int]]):
        self.luts[lut_name] = lut
        self.var_exponents[lut_name] = exponent

    def print(self):
        print(self.var_exponents)
        print(self.var_layout)
        print(self.var_permute)
        print(self.var_config)



import json
import os
from typing import List

from esp_ppq.core import NetworkFramework
from esp_ppq.IR import BaseGraph, GraphExporter

from .onnx_exporter import OnnxExporter


class MNNExporter(GraphExporter):
    def export_onnx_quantization_config(self, config_path: str, graph: BaseGraph):
        quant_info_json = {}
        shape = {}
        op_tensor_scales = []
        op_tensor_names = []
        for input_name in graph.inputs.keys():
            quant_var = graph.inputs[input_name]
            shape["channels"] = quant_var.shape[1]
            shape["height"] = quant_var.shape[2]
            shape["width"] = quant_var.shape[3]
        quant_info_json["shape"] = shape

        for op in graph.topological_sort():
            if op.type in {"Conv", 'Add'}:
                op_tensor_scales.clear()
                op_tensor_names.clear()
                for cfg, var in op.config_with_variable:
                    if not cfg.can_export(export_overlapped=True):
                        continue
                    if var.is_parameter:
                        continue
                    op_tensor_scales.append(cfg.scale.item())
                    op_tensor_names.append(var.name)

                assert len(op_tensor_scales) == len(op_tensor_names)
                if op.type == "Conv":
                    base_name = op_tensor_names[1]
                    input_tensor_name = base_name + " input_tensor_0"
                    output_tensor_name = base_name + " output_tensor_0"
                    quant_info_json[input_tensor_name] = op_tensor_scales[0]
                    quant_info_json[output_tensor_name] = op_tensor_scales[1]
                if op.type == "Add":
                    base_name = op_tensor_names[2]
                    output_tensor_name = base_name + " output_tensor_0"
                    quant_info_json[output_tensor_name] = op_tensor_scales[1]
        json_qparams_str = json.dumps(quant_info_json, indent=4)
        with open(config_path, "w") as json_file:
            json_file.write(json_qparams_str)

    def export(
        self,
        file_path: str,
        graph: BaseGraph,
        config_path: str = None,
        input_shapes: List[List[int]] = [[1, 3, 224, 224]],
    ):
        if graph._built_from == NetworkFramework.ONNX:
            if config_path is not None:
                self.export_onnx_quantization_config(config_path, graph)
            exporter = OnnxExporter()
            exporter.export(file_path=file_path, graph=graph, config_path=None)



from typing import Union

import torch

from esp_ppq.core import (
    PASSIVE_OPERATIONS,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
)
from esp_ppq.IR import BaseGraph, Operation

from .base import BaseQuantizer


class GraphCoreQuantizer(BaseQuantizer):
    def __init__(self, graph: BaseGraph) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._exponent_bits = 4
        self._quant_min = -448.0
        self._quant_max = +448.0

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=self._exponent_bits,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='floating',
        )

        # 一些特殊的算子需要更复杂的量化逻辑，我们在下面的代码中对量化信息进行调整
        if operation.type in {'Conv', 'ConvTranspose', 'Gemm', 'MatMul'}:
            # 将 Conv 的参数设置为对称 per-channel 量化
            if operation.type in {'Conv', 'ConvTranspose'}:
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL
                    + QuantizationProperty.PER_CHANNEL
                    + QuantizationProperty.FLOATING
                    + QuantizationProperty.POWER_OF_2
                )
                # 量化的 channel_axis 对于 conv 而言是 0，对于反卷积而言是 1
                # 参数量化使用 min max observe
                conv_weight_config.channel_axis = 1 if operation.type == 'ConvTranspose' else 0
                conv_weight_config.observer_algorithm = 'floating'

            # 将 Gemm 的参数设置为对称 per-channel 量化
            elif operation.type in {'Gemm', 'MatMul'}:
                gemm_weight_config = base_quant_config.input_quantization_config[1]
                gemm_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL
                    + QuantizationProperty.PER_CHANNEL
                    + QuantizationProperty.FLOATING
                    + QuantizationProperty.POWER_OF_2
                )
                gemm_weight_config.channel_axis = 0
                gemm_weight_config.observer_algorithm = 'floating'

            # 如果有 bias, 那 bias 就不量化了，因为我也不知道应该怎么量化他
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32

        if operation.type in PASSIVE_OPERATIONS:
            # 有一些算子是被动量化的，他们的输入输出将共享量化 scale
            base_quant_config.is_active_quant_op = False

        return base_quant_config

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.PPL_CUDA_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quant_operation_types(self) -> set:
        return {
            'Conv',
            'Relu',
            'PRelu',
            'Clip',
            'Gemm',
            'Resize',
            'MaxPool',
            'AveragePool',
            'GlobalMaxPool',
            'GlobalAveragePool',
            'Mul',
            'Add',
            'LeakyRelu',
            'Split',
            'Concat',
            'Transpose',
            'Slice',
            'Reshape',
            'Flatten',
            'Sigmoid',
            'ReduceMean',
            'ConvTranspose',
            'MatMul',
        }

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL
            + QuantizationProperty.PER_TENSOR
            + QuantizationProperty.FLOATING
            + QuantizationProperty.POWER_OF_2
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip', 'Sigmoid', 'LeakyRelu', 'Swish', 'Mish', 'PRelu'}


class TensorRTQuantizer_FP8(BaseQuantizer):
    def __init__(self, graph: BaseGraph) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._exponent_bits = 4
        self._quant_min = -448.0
        self._quant_max = +448.0

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=self._exponent_bits,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='floating',
        )

        # 一些特殊的算子需要更复杂的量化逻辑，我们在下面的代码中对量化信息进行调整
        if operation.type in {'Conv', 'ConvTranspose', 'Gemm', 'MatMul'}:
            # 将 Conv 的参数设置为对称 per-channel 量化
            if operation.type in {'Conv', 'ConvTranspose'}:
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL
                    + QuantizationProperty.PER_CHANNEL
                    + QuantizationProperty.FLOATING
                    + QuantizationProperty.POWER_OF_2
                )
                # 量化的 channel_axis 对于 conv 而言是 0，对于反卷积而言是 1
                # 参数量化使用 min max observe
                conv_weight_config.channel_axis = 1 if operation.type == 'ConvTranspose' else 0
                conv_weight_config.observer_algorithm = 'floating'

            # 将 Gemm 的参数设置为对称 per-channel 量化
            elif operation.type in {'Gemm', 'MatMul'}:
                gemm_weight_config = base_quant_config.input_quantization_config[1]
                gemm_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL
                    + QuantizationProperty.PER_CHANNEL
                    + QuantizationProperty.FLOATING
                    + QuantizationProperty.POWER_OF_2
                )
                gemm_weight_config.channel_axis = 0
                gemm_weight_config.observer_algorithm = 'floating'

            # 如果有 bias, 那 bias 就不量化了，因为我也不知道应该怎么量化他
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32

            # 所有算子只量化输入
            for output_config in base_quant_config.output_quantization_config:
                output_config.state = QuantizationStates.FP32

        if operation.type in PASSIVE_OPERATIONS:
            # 有一些算子是被动量化的，他们的输入输出将共享量化 scale
            base_quant_config.is_active_quant_op = False

        return base_quant_config

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.PPL_CUDA_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quant_operation_types(self) -> set:
        return {'Conv', 'Gemm', 'AveragePool', 'GlobalAveragePool', 'ConvTranspose', 'MatMul'}

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL
            + QuantizationProperty.PER_TENSOR
            + QuantizationProperty.FLOATING
            + QuantizationProperty.POWER_OF_2
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip', 'Sigmoid', 'LeakyRelu', 'Swish', 'Mish', 'PRelu'}



import os
from typing import List

from esp_ppq.core import DataType, NetworkFramework, QuantizationProperty, QuantizationStates, ppq_warning
from esp_ppq.IR import BaseGraph, GraphExporter

from .onnx_exporter import OnnxExporter
from .util import convert_value

ASCEND_QUANT_OP = {"Conv", "ConvTranspose", "Gemm", "AveragePool"}

FLT_EPSILON = 1.1920929e-7


def adapt_scale(op, scale):
    min = FLT_EPSILON
    max = 1.0 / FLT_EPSILON
    if scale < min:
        scale = FLT_EPSILON
        ppq_warning(f'{op.name} scale is too small: {scale}.')
    elif scale > max:
        scale = max
        ppq_warning(f'{op.name} scale is too large: {scale}.')
    return scale


def check_offset(offset):
    if offset > 127 or offset < -128:
        raise RuntimeError(f'This offset value {offset} does not belong to the range [-128,127].')


def generate_shape(shape):
    channels = ""
    height = ""
    width = ""
    if len(shape) == 2:
        channels = shape[1]
        height = 1
        width = 1
    elif len(shape) == 4:
        _, channels, height, width = shape
    else:
        raise RuntimeError('Please design this shape yourself.')
    return channels, height, width


class AscendExporter(GraphExporter):
    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        new_config_path = os.path.splitext(config_path)[0] + ".txt"
        matched_nodes = []

        for op in graph.topological_sort():
            if op.type in {"Conv", "ConvTranspose", "Gemm"}:
                if not hasattr(op, 'config'):
                    ppq_warning(f'This op does not write quantization parameters: {op.name}.')
                    continue
                op_name = '''\"''' + op.name + '''\"'''
                quant_unit_list = []
                quant_unit_list.append("record {\n")
                quant_unit_list.append("  key: " + op_name + "\n")
                quant_unit_list.append("  value {\n")

                input_cfg = op.config.input_quantization_config[0]
                assert input_cfg.state == QuantizationStates.ACTIVATED and input_cfg.policy.has_property(
                    QuantizationProperty.PER_TENSOR
                )

                scale_d = input_cfg.scale.item()
                offset_d = int(input_cfg.offset.item()) - 128

                check_offset(offset_d)
                quant_unit_list.append("    scale_d: " + str(adapt_scale(op, scale_d)) + "\n")
                quant_unit_list.append("    offset_d: " + str(offset_d) + "\n")

                weight_config = op.config.input_quantization_config[1]
                scale_list = convert_value(weight_config.scale, False, DataType.FP32)

                if op.type == "Gemm":
                    assert isinstance(scale_list, float), 'Gemm can only have one scale.'
                    scale_list = [scale_list]

                for scale_w in scale_list:
                    quant_unit_list.append("    scale_w: " + str(adapt_scale(op, scale_w)) + "\n")

                for _ in range(len(scale_list)):
                    quant_unit_list.append("    offset_w: " + "0" + "\n")

                channels, height, width = generate_shape(op.inputs[0].shape)

                quant_unit_list.append("    channels: " + str(channels) + "\n")
                quant_unit_list.append("    height: " + str(height) + "\n")
                quant_unit_list.append("    width: " + str(width) + "\n")
                quant_unit_list.append("  }" + "\n")
                quant_unit_list.append("}" + "\n")
                matched_nodes.append(quant_unit_list)

            elif op.type == "AveragePool":
                if not hasattr(op, 'config'):
                    ppq_warning(f'This op does not write quantization parameters: {op.name}.')
                    continue
                op_name = '''\"''' + op.name + '''\"'''
                quant_unit_list = []
                quant_unit_list.append("record {\n")
                quant_unit_list.append("  key: " + op_name + "\n")
                quant_unit_list.append("  value {\n")
                input_cfg = op.config.input_quantization_config[0]
                assert input_cfg.state == QuantizationStates.ACTIVATED and input_cfg.policy.has_property(
                    QuantizationProperty.PER_TENSOR
                )
                scale_d = input_cfg.scale.item()

                offset_d = int(input_cfg.offset.item()) - 128
                check_offset(offset_d)
                quant_unit_list.append("    scale_d: " + str(adapt_scale(op, scale_d)) + "\n")
                quant_unit_list.append("    offset_d: " + str(offset_d) + "\n")
                _, channels, height, width = op.inputs[0].shape
                quant_unit_list.append("    channels: " + str(channels) + "\n")
                quant_unit_list.append("    height: " + str(height) + "\n")
                quant_unit_list.append("    width: " + str(width) + "\n")
                quant_unit_list.append("  }" + "\n")
                quant_unit_list.append("}" + "\n")
                matched_nodes.append(quant_unit_list)
            else:
                continue
        fd = open(new_config_path, 'w+')
        for tem_list in matched_nodes:
            for tem_str in tem_list:
                fd.write(tem_str)
        fd.close()

    def export(
        self,
        file_path: str,
        graph: BaseGraph,
        config_path: str = None,
        input_shapes: List[List[int]] = [[1, 3, 224, 224]],
    ):
        if config_path is not None:
            self.export_quantization_config(config_path, graph)

        _, ext = os.path.splitext(file_path)
        if ext == '.onnx' or graph._built_from == NetworkFramework.ONNX:
            exporter = OnnxExporter()
            exporter.export(file_path=file_path, graph=graph, config_path=None)
